"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const tokens_1 = require("../../tokens");
const createSqlTag_1 = require("../createSqlTag");
const ava_1 = __importDefault(require("ava"));
const roarr_1 = require("roarr");
const test = ava_1.default;
const sql = (0, createSqlTag_1.createSqlTag)();
test.beforeEach((t) => {
    t.context.logs = [];
    roarr_1.ROARR.write = (message) => {
        t.context.logs.push(JSON.parse(message));
    };
});
test('creates an object describing a query', (t) => {
    const query = sql.fragment `SELECT 1`;
    t.deepEqual(query, {
        sql: 'SELECT 1',
        type: tokens_1.FragmentToken,
        values: [],
    });
});
test('creates an object describing query value bindings', (t) => {
    const query = sql.fragment `SELECT ${'foo'}`;
    t.deepEqual(query, {
        sql: 'SELECT $slonik_1',
        type: tokens_1.FragmentToken,
        values: ['foo'],
    });
});
test('creates an object describing query value bindings (multiple)', (t) => {
    const query = sql.fragment `SELECT ${'foo'}, ${'bar'}`;
    t.deepEqual(query, {
        sql: 'SELECT $slonik_1, $slonik_2',
        type: tokens_1.FragmentToken,
        values: ['foo', 'bar'],
    });
});
test('nests sql templates', (t) => {
    const query0 = sql.fragment `SELECT ${'foo'} FROM bar`;
    const query1 = sql.fragment `SELECT ${'baz'} FROM (${query0})`;
    t.deepEqual(query1, {
        sql: 'SELECT $slonik_1 FROM (SELECT $slonik_2 FROM bar)',
        type: tokens_1.FragmentToken,
        values: ['baz', 'foo'],
    });
});
test('throws if bound an undefined value', (t) => {
    const error = t.throws(() => {
        // @ts-expect-error - intentional
        sql.fragment `SELECT ${undefined}`;
    });
    t.is(error?.message, 'SQL tag cannot be bound to undefined value at index 1.');
});
// eslint-disable-next-line n/no-process-env
if (process.env.ROARR_LOG === '1') {
    // This test is only valid when ROARR_LOG=1.
    // TODO find a way to test this without relying on environment variables.
    test.serial('logs all bound values if one is undefined', (t) => {
        t.throws(() => {
            // @ts-expect-error - intentional
            sql.fragment `SELECT ${undefined}`;
        });
        const targetMessage = t.context.logs.find((message) => {
            return message.message === 'bound values';
        });
        t.truthy(targetMessage);
        t.deepEqual(targetMessage.context.parts, ['SELECT ', '']);
    });
}
test('the sql property is immutable', (t) => {
    const query = sql.fragment `SELECT 1`;
    t.throws(() => {
        // @ts-expect-error This is intentional.
        query.sql = 'SELECT 2';
    });
});
/**
 * https://github.com/gajus/slonik/pull/552
 */
test('copes with dollar-number in table name', (t) => {
    const query0 = sql.fragment `discounted_to_$1 (offer_id INTEGER)`;
    const query1 = sql.fragment `CREATE TABLE ${query0}`;
    t.deepEqual(query1, {
        sql: 'CREATE TABLE discounted_to_$1 (offer_id INTEGER)',
        type: tokens_1.FragmentToken,
        values: [],
    });
});
/**
 * https://github.com/gajus/slonik/pull/552
 */
test('copes with dollar-number in column name (CREATE TABLE)', (t) => {
    const query0 = sql.fragment `offers (discounted_to_$1 BOOLEAN)`;
    const query1 = sql.fragment `CREATE TABLE ${query0}`;
    t.deepEqual(query1, {
        sql: 'CREATE TABLE offers (discounted_to_$1 BOOLEAN)',
        type: tokens_1.FragmentToken,
        values: [],
    });
});
/**
 * https://github.com/gajus/slonik/pull/552
 */
test('copes with dollar-number in column name (SELECT)', (t) => {
    const query0 = sql.fragment `"discounted_to_$1" IS TRUE`;
    const query1 = sql.fragment `SELECT * FROM offers WHERE ${query0}`;
    t.deepEqual(query1, {
        sql: 'SELECT * FROM offers WHERE "discounted_to_$1" IS TRUE',
        type: tokens_1.FragmentToken,
        values: [],
    });
});
/**
 * https://github.com/gajus/slonik/pull/552
 */
test('copes with dollar-number in function definitions', (t) => {
    // example function from https://www.postgresql.org/docs/current/sql-createfunction.html
    const query0 = sql.fragment `add(integer, integer) RETURNS integer
      AS 'select $1 + $2;'
      LANGUAGE SQL
      IMMUTABLE
      RETURNS NULL ON NULL INPUT`;
    const query1 = sql.fragment `CREATE FUNCTION ${query0}`;
    t.deepEqual(query1, {
        sql: `CREATE FUNCTION add(integer, integer) RETURNS integer
      AS 'select $1 + $2;'
      LANGUAGE SQL
      IMMUTABLE
      RETURNS NULL ON NULL INPUT`,
        type: tokens_1.FragmentToken,
        values: [],
    });
});
//# sourceMappingURL=sql.test.js.map